// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#ifndef HPPWALKFOOTPLANNER_ROBOPTIM_OPTIMIZER_HH
# define HPPWALKFOOTPLANNER_ROBOPTIM_OPTIMIZER_HH
# include <KineoWorks2/kwsPathOptimizer.h>
# include <KineoWorks2/kwsDevice.h>

# include <roboptim/trajectory/fwd.hh>
# include <roboptim/trajectory/cubic-b-spline.hh>

# include <hpp/core/fwd.hh>
# include <hpp/walkfootplanner/fwd.hh>
# include <hpp/walkfootplanner/spline-directpath.hh>
# include <hpp/walkfootplanner/config.hh>

namespace hpp
{
  namespace walkFootPlanner
  {
    /**
       \brief Path optimiser using numerical optimization techniques

       The input path should be composed of CkwsPlusDPLinear direct paths.
    */
    class HPP_WALKFOOTPLANNER_DLLAPI RobOptimOptimizer : public CkwsPathOptimizer
    {
    public:
      typedef SplineDirectPath::controlPointsFittingStrategy_t
	controlPointsFittingStrategy_t;
      virtual ~RobOptimOptimizer();
      /**
	 \brief Create an object and return a shared pointer.
      */
      static RobOptimOptimizerShPtr create(double minDistance,
					   double stepSpace,
					   double vMaxX,
					   double vMaxY,
					   double vMaxTheta,
					   core::Planner *inHppPlanner = 0,
					   controlPointsFittingStrategy_t strategy
					   = &defaultControlPointsFittingStrategy);

      void vMaxX (const double vMaxX) throw ()
      {
	vMaxX_ = vMaxX;
      }

      void vMaxY (const double vMaxY) throw ()
      {
	vMaxY_ = vMaxY;
      }

      void vMaxTheta (const double vMaxTheta) throw ()
      {
	vMaxTheta_ = vMaxTheta;
      }

    protected:
      /**
	 \brief Constructor
      */
      RobOptimOptimizer(double minDistance,
			double stepSpace,
			double vMaxX,
			double vMaxY,
			double vMaxTheta,
			core::Planner *inHppPlanner,
			controlPointsFittingStrategy_t strategy
			= &defaultControlPointsFittingStrategy);

      /**
	 \brief Initialization
      */
      ktStatus init(RobOptimOptimizerWkPtr inWeakPtr);

      /**
	 \brief Optimisation  of the path
      */
      ktStatus doOptimizePath(const CkwsPathShPtr &inOutPath);

      /**
	 \brief Compute the initial spline from input path

	 \param inPath input path computed by KineoWorks
	 \return a \f$C^2\f$ spline fitting the input path

	 \note The input path is assumed to be a concatenation of linear direct paths.
	 Therefore, the first two derivative of the spline at connection configurations
	 should be equalt to 0.
      */
      roboptim::CubicBSpline* computeSplineFromPath(const CkwsPathConstShPtr& inPath) const;

      virtual double completion() const
      {
	return attCompletion;
      }

      virtual ktStatus doComputeCompletion (double &o_completion) const
      {
	o_completion = attCompletion;
	return KD_OK;
      }

    private:
      /// \brief Store weak pointer to itself.
      RobOptimOptimizerWkPtr attWeakPtr;

      double attMinDistance;
      double attCompletion;
      double attStepSpace;

      double vMaxX_;
      double vMaxY_;
      double vMaxTheta_;

      /// \brief Optionally store a core::Planner object.
      core::Planner* attHppPlanner;

      /// \brief Strategy for control points fitting when instantiating
      /// a spline direct path.
      controlPointsFittingStrategy_t strategy_;
    };
  } // end of namespace walkFootPlanner.
} // end of namespace hpp.

#endif //! HPP_WALKFOOTPLANNER_ROBOPTIM_OPTIMIZER_HH
