// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <boost/numeric/conversion/converter.hpp>

#include <jrl/mathtools/angle.hh>

#include <KineoWorks2/kwsConfig.h>
#include <KineoWorks2/kwsPath.h>

#include "hpp/util/debug.hh"
#include "hpp/util/indent.hh"
#include "hpp/walkfootplanner/io.hh"
#include "hpp/walkfootplanner/footstep-sequence.hh"
#include "hpp/walkfootplanner/footstep-sequence-builder.hh"
#include "hpp/walkfootplanner/planner.hh"


namespace hpp
{
  namespace walkFootPlanner
  {
    HPP_WALKFOOTPLANNER_DLLAPI footStep_t
    convertToFootStepXYTheta (CkwsConfigShPtr cfg)
    {
      static const unsigned DOF_X = 0;
      static const unsigned DOF_Y = 1;
      static const unsigned DOF_RZ = 5;

      footStep_t footPrint;
      footPrint[0] = cfg->dofValue (DOF_X);
      footPrint[1] = cfg->dofValue (DOF_Y);
      footPrint[2] = cfg->dofValue (DOF_RZ);
      return footPrint;
    }

    namespace
    {
      static bool
      leftFootFirst (FootStepSequence& footStepSequence,
		     CkwsPathShPtr path,
		     double stepLength,
		     FootStepSequenceBuilder::footStepGetter_t footStepGetter)
      {
	assert (footStepGetter);
	const footStep_t& init = footStepSequence.initialFootStep ();
	footStep_t pos = footStepGetter (path->configAtDistance (stepLength));

	const double res =
	  std::cos(init[2]) * (pos[1] - init[1])
	  - std::sin (init[2]) * (pos[0] - init[0]);

	return res > 0;
      }

      static void
      addParallelFootStep (FootStepSequence& footStepSequence,
			   const Planner& planner,
			   int& supportFoot,
			   bool absolute)
      {
	footStep_t footPrint;
	footPrint[0] = 0.;
	if (absolute)
	  footPrint[1] = supportFoot * planner.stepSpacing () / 2.;
	else
	  footPrint[1] = supportFoot * planner.stepSpacing ();
	footPrint[2] = 0.;
	footStepSequence.footSteps ().push_back (footPrint);
	supportFoot *= -1;
      }



      static footStep_t
      convertToPgFrame (const footStep_t& current, const footStep_t& init)
      {
	footStep_t footPrint;
	const double cos = std::cos (init[2]);
	const double sin = std::sin (init[2]);

	footPrint[0] =
	  (current[0] - init[0]) * cos
	  + (current[1] - init[1]) * sin;
	footPrint[1] =
	  - (current[0] - init[0]) * sin
	  + (current[1] - init[1]) * cos;

	jrlMathTools::Angle theta = current[2] - init[2];
	footPrint[2] = theta;

	return footPrint;
      }

      static footStep_t
      getNewAbsoluteFootStep (const footStep_t& currentPgFrame,
			      const Planner& planner,
			      int supportFoot,
			      double epsilon)
      {
	footStep_t newAbsolute = currentPgFrame;
	const double l = (planner.stepSpacing () + epsilon) / 2.;

	newAbsolute[0] += l * std::sin (currentPgFrame[2]) * -supportFoot;
	newAbsolute[1] += l * std::cos (currentPgFrame[2]) * supportFoot;
	return newAbsolute;
      }

      // compute x, y ,theta Relatif of the new feet from the old feet
      // WARNING : express xNewRelatif , yNewRelatif in the Old foot frame
      // with the orientation of the New foot
      // S=[XNewRelatif, yNewRelatif, thetaNewRelatif] ,
      // p=[xNewAbsolu, yNewAbsolu], pOld=[xOld, yOld],
      // R(dtheta+thetaOld)= rotation matrix
      // R * S = (p - pOld)
      // BECAUSE : WPG first compute the rotation of the foot then the
      // translation.
      static footStep_t
      convertStep (const footStep_t& newAbsolute, const footStep_t& oldAbsolute)
      {
	footStep_t pgStep;
	const double cos = std::cos (newAbsolute[2]);
	const double sin = std::sin (newAbsolute[2]);

	pgStep[0] =
	  (newAbsolute[0] - oldAbsolute[0]) * cos
	  + (newAbsolute[1] - oldAbsolute[1]) * sin;
	pgStep[1] =
	  - (newAbsolute[0] - oldAbsolute[0]) * sin
	  + (newAbsolute[1] - oldAbsolute[1]) * cos;

	jrlMathTools::Angle theta = newAbsolute[2] - oldAbsolute[2];
	pgStep[2] = theta;
	pgStep[2] *= 180. / M_PI; // Degrees are required.
	return pgStep;
      }
    } // end of anonymous namespace.


    FootStepSequenceBuilder::FootStepSequenceBuilder (footStepGetter_t footStepGetter)
      : footStepGetter_ (footStepGetter)
    {
      assert (footStepGetter);
    }

    FootStepSequenceBuilder::~FootStepSequenceBuilder ()
    {}

    void
    FootStepSequenceBuilder::operator()
      (CkwsPathShPtr path, FootStepSequence& footStepSequence,
       const Planner& planner)
    {
      static const double epsilonInc = 0.01;
      static const int LEFT = -1;
      static const int RIGHT = 1;

      const double l = std::min (path->length (), planner.stepLength ());

      hppDout(info, "Path length: " << path->length ());
      hppDout(info, "Step length: " << planner.stepLength ());

      assert (planner.stepLength () > 0.);
      assert (path->length () >= 0.);

      using boost::numeric::converter;
      unsigned nbFullStep = converter<unsigned, double>::convert
	(path->length () / planner.stepLength ());
      hppDout(info, nbFullStep << " step(s) will be generated");

      // Box initial position.
      // The pattern generator starts from config (0,0,0).  We need to
      // move the box path by the inverse of the initial
      // configuration.
      footStepSequence.initialFootStep () =
	footStepGetter_ (path->configAtStart ());
      hppDout(info, "initial position: "
	      << footStepSequence.initialFootStep ());

      footStepSequence.leftFootFirst () =
	leftFootFirst (footStepSequence, path, l, footStepGetter_);
      hppDout (info, "starting with "
	       << (footStepSequence.leftFootFirst () ? "left" : "right")
	       << " foot");

      int supportFoot = footStepSequence.leftFootFirst () ? LEFT : RIGHT;

      hppDout (info, "removing previous footsteps");
      footStepSequence.footSteps ().clear ();

      addParallelFootStep (footStepSequence, planner, supportFoot, true);

      double distanceFromStart = 0;
      footStep_t oldAbsolute = footStepSequence.footSteps ()[0];

#ifndef HPP_GLOBAL_FRAME
      static const unsigned additionalStep = 0;
#else
      static const unsigned additionalStep = 1;
#endif

      for (unsigned i = 0; i < nbFullStep + additionalStep + 1; ++i)
	{
	  hppDout (info, "generating step " << i);
	  CkwsConfigShPtr cfg;

	  distanceFromStart += planner.stepLength ();
	  hppDout (info, "distance from starting point: " << distanceFromStart);

	  footStep_t current;
	  if (i >= nbFullStep)
	    current = footStepGetter_ (path->configAtEnd ());
	  else
	    current = footStepGetter_ (path->configAtDistance (distanceFromStart));
	  hppDout (info, "current footstep: " << current);

	  // replace the point in the PatternGenerator frame
	  footStep_t currentPgFrame = convertToPgFrame
	    (current, footStepSequence.initialFootStep ());
	  hppDout (info, "current footstep (PG frame): " << currentPgFrame);

	  // make sure there is enough space between feet
	  footStep_t newAbsolute;
	  double epsilon = 0.;
	  const double footSecurityDistanceSquare =
	    planner.footSecurityDistance () * planner.footSecurityDistance ();
	  double distanceBetweenFeetSquare = 0.;
	  do
	    {
	      newAbsolute = getNewAbsoluteFootStep
		(currentPgFrame, planner, supportFoot, epsilon);
	      const footStep_t delta = newAbsolute - oldAbsolute;
	      distanceBetweenFeetSquare =
		delta[0] * delta[0] + delta[1] * delta[1];
	      epsilon += epsilonInc;
	    }
	  while (distanceBetweenFeetSquare < footSecurityDistanceSquare);
	  epsilon -= epsilonInc;

	  hppDout (info, "epsilon: " << epsilon);
	  hppDout (info, "final footstep: " << newAbsolute);

	  footStep_t pgStep = newAbsolute;
#ifndef HPP_GLOBAL_FRAME
	  // convert to PG frame and express angle in degrees before
	  // pushing the step.
	  pgStep = convertStep (newAbsolute, oldAbsolute);
#endif
	  footStepSequence.footSteps ().push_back (pgStep);
	  hppDout (info, "pushed step: " << pgStep);
	  supportFoot *= -1;

	  oldAbsolute = newAbsolute;
	}

#ifndef HPP_GLOBAL_FRAME
      hppDout (info, "add last step (parallel feet)");
      addParallelFootStep (footStepSequence, planner, supportFoot, false);
#endif
    }

  } // end of namespace walkFootPlanner.
} // end of namespace hpp.
