// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <fstream>

#include <KineoWorks2/kwsJoint.h>

#include <hpp/model/body.hh>
#include <hpp/model/device.hh>

#include <roboptim/core/derivable-function.hh>
#include <roboptim/core/solver-factory.hh>
#include <roboptim/core/solver.hh>

#include <roboptim/core/visualization/gnuplot.hh>
#include <roboptim/trajectory/visualization/trajectory.hh>

#include <hpp/util/debug.hh>
#include <hpp/util/indent.hh>
#include <hpp/util/timer.hh>
#include "hpp/walkfootplanner/roboptim-optimizer.hh"
#include "hpp/walkfootplanner/spline-directpath.hh"
#include "hpp/walkfootplanner/minimize-time.hh"


#include <hpp/core/planner.hh>
#include "config.h"

// Namespaces.
using namespace roboptim;
using namespace roboptim::visualization;
using namespace roboptim::visualization::gnuplot;

namespace hpp
{
  namespace walkFootPlanner
  {
    RobOptimOptimizer::~RobOptimOptimizer()
    {
    }

    RobOptimOptimizerShPtr
    RobOptimOptimizer::create(double minDistance,
			      double stepSpace,
			      double vMaxX,
			      double vMaxY,
			      double vMaxTheta,
			      core::Planner *inHppPlanner,
			      controlPointsFittingStrategy_t strategy)
    {
      RobOptimOptimizer* optimizer =
	new RobOptimOptimizer (minDistance, stepSpace,
			       vMaxX, vMaxY, vMaxTheta,
			       inHppPlanner, strategy);
      RobOptimOptimizerShPtr optimizerShPtr (optimizer);

      if (optimizer->init(optimizerShPtr) != KD_OK)
	optimizerShPtr.reset();
      return optimizerShPtr;
    }

    RobOptimOptimizer::RobOptimOptimizer(double minDistance,
					 double stepSpace,
					 double vMaxX,
					 double vMaxY,
					 double vMaxTheta,
					 core::Planner* inHppPlanner,
					 controlPointsFittingStrategy_t strategy)
      : CkwsPathOptimizer(),
	attWeakPtr (),
	attMinDistance (minDistance),
	attCompletion (0.),
	attStepSpace (stepSpace),
	vMaxX_ (vMaxX),
	vMaxY_ (vMaxY),
	vMaxTheta_ (vMaxTheta),
	attHppPlanner(inHppPlanner),
	strategy_ (strategy)
    {
    }


    ktStatus RobOptimOptimizer::init(RobOptimOptimizerWkPtr inWeakPtr)
    {
      if (CkwsPathOptimizer::init(inWeakPtr) != KD_OK) {
	return KD_ERROR;
      }
      attWeakPtr = inWeakPtr;
      return KD_OK;
    }


    namespace
    {
      void plotTrajectory (const CubicBSpline& before,
			   const CubicBSpline& after)
      {
	Gnuplot gnuplot = Gnuplot::make_interactive_gnuplot ();
	gnuplot
	  << gnuplot::set ("multiplot layout 1,2")
	  << gnuplot::set ("grid")
	  << plot_xy (before)
	  << plot_xy (after)
	  << unset ("multiplot");
	std::ofstream debugstream
	  (debug::getFilename ("trajectory.gp", PACKAGE_TARNAME).c_str ());
	debugstream << gnuplot;
      }


      // Update path if the time is optimized.
      ktStatus updatePath (const CkwsPathShPtr& inOutPath,
			   const CubicBSpline& spline,
			   const Function::vector_t& params,
			   const MinimizeTime& problem)
      {
	hppDout (info, "Optimize time.");

	// Convert raw parameters into a free time trajectory.
	FreeTimeTrajectory<CubicBSpline> fttTraj (spline, 1.);
	fttTraj.setParameters (params);

	assert (fttTraj.parameters ().size () ==
		spline.parameters ().size () + 1);
	Function::vector_t newParams =
	  removeScaleFromParameters (fttTraj.parameters ());

	// Make updated spline with new time interval.
	CubicBSpline updatedSpline
	  (fttTraj.timeRange (), fttTraj.outputSize (), newParams);

	// Create a direct path from the updated spline.
	SplineDirectPathShPtr optimizedDirectPath
	  (SplineDirectPath::create(updatedSpline, inOutPath->device()));
	if (!optimizedDirectPath)
	  return KD_ERROR;

	// Plot problem and trajectory.
	SplineDirectPathShPtr directPath =
	  SplineDirectPath::create(spline, inOutPath->device());
	std::ofstream speedBefore
	  (debug::getFilename
	   ("constraints-before.gp", PACKAGE_TARNAME).c_str ());
	std::ofstream speedAfter
	  (debug::getFilename
	   ("constraints-after.gp", PACKAGE_TARNAME).c_str ());
	speedBefore << problem.display (directPath);
	speedAfter << problem.display (optimizedDirectPath);

	plotTrajectory (spline, updatedSpline);

	// Update the path.
	inOutPath->clear ();
	if (inOutPath->appendDirectPath (optimizedDirectPath) != KD_OK)
	  return KD_ERROR;
	return KD_OK;
      }
    }

    ktStatus RobOptimOptimizer::doOptimizePath(const CkwsPathShPtr &inOutPath)
    {
      /* First optimize with adaptive shortcut */
      CkwsDistanceConstShPtr distance = CkwsDistance::create ();

      // Get distance function associated to robot if any
      CkwsDeviceShPtr hppRobot = inOutPath->device ();
      if (hppRobot && hppRobot->distance ())
	distance = hppRobot->distance ();

      CkwsLoopOptimizerShPtr pathOptimizer =
	CkwsRandomOptimizer::create();
      pathOptimizer->distance(distance);
      pathOptimizer->shortcutMethod(CkwsShortcutDirect::create());
      pathOptimizer->maxNbLoop(50);

      pathOptimizer->optimizePath(inOutPath);

      if (attHppPlanner)
	attHppPlanner->addPath
	  (0, KIT_DYNAMIC_PTR_CAST (CkwsPath, inOutPath->clone ()));
      else
	hppDout (info, "no core::Planner provided.");

      hppDout (info, "Start optimization with the following parameters:"
	       << incindent << iendl
	       << "vMaxX = " << vMaxX_ << iendl
	       << "vMaxY = " << vMaxY_ << iendl
	       << "vMaxTheta = " << vMaxTheta_ << decindent);

      // Define the Spline.
      SplineDirectPathShPtr directPath =
	SplineDirectPath::create(inOutPath, strategy_);
      if (!directPath)
	return KD_ERROR;
      CubicBSpline spline = directPath->spline ();

      // Get robot body.
      model::BodyShPtr body = KIT_DYNAMIC_PTR_CAST
	(model::Body, directPath->device ()->rootJoint ()->attachedBody ());
      assert (!!body);

      // Optimize the Spline.
      typedef MinimizeTime problemBuilder_t;
      typedef problemBuilder_t::solver_t solver_t;

      problemBuilder_t problemBuilder (spline, body, directPath,
				       attMinDistance, attStepSpace,
				       vMaxX_, vMaxY_, vMaxTheta_);
      const problemBuilder_t::problem_t& problem =
	problemBuilder.getProblem ();

#ifdef HPP_DEBUG
      // Get the starting point
      problemBuilder_t::problem_t::startingPoint_t startPoint =
	problem.startingPoint();
      const Function::vector_t& initialGuess = *startPoint;
#endif // HPP_DEBUG

      solver_t::result_t res;
      ktStatus status = KD_ERROR;
      try
	{
	  SolverFactory<solver_t> factory ("cfsqp", problem);
	  solver_t& solver = factory ();

	  // Set maximum number of iterations.
	  solver.parameters ()["max-iterations"].value = 50;
	  // Set precision
	  solver.parameters ()["cfsqp.Eps"].value = 1e-4;
	  solver.parameters ()["cfsqp.Epseqn"].value = 1e-4;
	  // Do not display any debugging information.
	  solver.parameters ()["cfsqp.iprint"].value = 0;

	  hppStartBenchmark (OPTIMIZATION);
	  solver.solve ();
	  hppStopBenchmark (OPTIMIZATION);
	  hppDisplayBenchmark (OPTIMIZATION);

	  hppDout (info, solver);

	  // Update the path.
	  switch (solver.minimumType ())
	    {
	    case GenericSolver::SOLVER_VALUE:
	      {
		Result result = solver.getMinimum<Result> ();
		hppDout (info, result);
		status = updatePath (inOutPath, spline, result.x, problemBuilder);
	      }
	      break;

	    case GenericSolver::SOLVER_VALUE_WARNINGS:
	      {
		ResultWithWarnings result =
		  solver.getMinimum<ResultWithWarnings> ();
		hppDout (info, result);
		status = updatePath (inOutPath, spline, result.x, problemBuilder);
	      }
	      break;

	    case GenericSolver::SOLVER_NO_SOLUTION:
	      {
		hppDout (error, "Solver did not find any solution.");
		status = KD_ERROR;
#ifdef HPP_DEBUG
		status = updatePath (inOutPath, spline, initialGuess,
				     problemBuilder);
#endif // HPP_DEBUG
		break;
	      }
	    case GenericSolver::SOLVER_ERROR:
	      {
		hppDout (error, solver.getMinimum<SolverError> ());
		status = KD_ERROR;
#ifdef HPP_DEBUG
		status = updatePath (inOutPath, spline, initialGuess,
				     problemBuilder);
#endif // HPP_DEBUG
		break;
	      }
	    }
	}
      catch (std::runtime_error& e)
	{
	  hppDout (error, "Can not load the plug-in: " << e.what ());
	  status = KD_ERROR;
	}

      attCompletion = 1.;
      return status;
    }
  } // end of namespace walkFootPlanner.
} // end of namespace hpp.

//  LocalWords: holonomic
