// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <iostream>
#include <boost/format.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <KineoWorks2/kwsDirectPath.h>
#include <KineoWorks2/kwsJoint.h>
#include <KineoWorks2/kwsPath.h>

#include <kwsioConfig.h>
#include <kwsioMat.h>

#include <hpp/model/joint.hh>
#include <hpp/model/body.hh>
#include <hpp/model/device.hh>

#include <roboptim/core/derivable-function.hh>
#include <roboptim/core/finite-difference-gradient.hh>
#include <roboptim/trajectory/free-time-trajectory.hh>

#include "hpp/util/debug.hh"
#include "hpp/util/indent.hh"
#include "hpp/walkfootplanner/robot-obstacle-distance.hh"
#include "hpp/walkfootplanner/spline-directpath.hh"


using namespace roboptim;

namespace hpp
{
  namespace walkFootPlanner
  {

    bool
    LtCkwsConfig::operator () (const CkwsConfig& l, const CkwsConfig& r)
    {
      unsigned ls = l.size ();
      unsigned rs = r.size ();

      if (ls != rs)
	return ls < rs;

      for (unsigned i = 0 ; i < ls; ++i)
	{
	  double lv = l.dofValue (i);
	  double rv = r.dofValue (i);
	  if (lv != rv)
	    return lv < rv;
	}
      return false;
    }

    DistanceProxy::DistanceProxy (unsigned obstacle,
				  model::DeviceShPtr device,
				  model::BodyShPtr body)
      : obstacle_ (obstacle),
	device_ (device),
	body_ (body)
    {}

    DistAndPairsOfPointsResult
    DistanceProxy::operator () (CkwsConfig& cfg)
    {
      cache_t::const_iterator it = cache_.find (cfg);
      if (it != cache_.end ())
	return it->second;

      // Set robot configuration.
      bool status = device_->hppSetCurrentConfig (cfg);
      if (!status)
	assert (0 && "Can not set configuration.");

      DistAndPairsOfPointsResult result;
      CkcdObjectShPtr objectBody;
      CkcdObjectShPtr objectEnv;

      ktStatus err = body_->distAndPairsOfPoints (obstacle_,
						  result.distance,
						  result.pointBody,
						  result.pointEnv);
      if (err != KD_OK)
	{
	  const char errorStr[] = "Failed to compute the object distance.";
	  hppDout (warning, errorStr);
	  assert (errorStr != errorStr);
	}
      cache_[cfg] = result;
      return result;
    }

    RobotObstacleDistance::RobotObstacleDistance (SplineDirectPathShPtr dp,
						  roboptim::StableTimePoint tpt,
						  size_type obstacleId)
      throw ()
      : DerivableFunction (dp->spline ().parameters ().size () + 1, 1,
			   (boost::format ("distance to obstacle %1% (%2%)")
			    % obstacleId
			    % tpt.getAlpha ()).str ()),
	dp_ (dp),
	tpt_ (tpt),
	obstacleId_ (obstacleId),
	hppDevice_ (KIT_DYNAMIC_PTR_CAST (model::Device, dp->device ())),
	hppBody_ (KIT_DYNAMIC_PTR_CAST
		  (model::Body, dp->device ()->rootJoint ()->attachedBody ())),
	cfg_ (dp->device ()),
	distanceProxy_ (obstacleId_, hppDevice_, hppBody_)
    {
    }

    RobotObstacleDistance::~RobotObstacleDistance () throw ()
    {
    }

    void
    RobotObstacleDistance::impl_compute (result_t& res, const argument_t& p) const throw ()
    {
      using namespace boost::numeric::ublas;

      res.clear ();

      typedef FreeTimeTrajectory<roboptim::CubicBSpline>
	freeTime_t;
      static freeTime_t fttTraj (dp_->spline (), 1.);
      fttTraj.setParameters (p);
      convertResult (fttTraj (tpt_), dp_->device (), cfg_);

      // Compute distance.
      DistAndPairsOfPointsResult result = distanceProxy_ (cfg_);
      res[0] = result.distance;
    }

    void
    RobotObstacleDistance::impl_gradient (gradient_t& grad, const argument_t& p,
					size_type i) const throw ()
    {
      assert (i == 0);
      using namespace boost::numeric::ublas;

      grad.clear ();

      typedef FreeTimeTrajectory<roboptim::CubicBSpline>
	freeTime_t;
      static freeTime_t fttTraj (dp_->spline (), 1.);
      fttTraj.setParameters (p);
      convertResult (fttTraj (tpt_), dp_->device (), cfg_);

      // Set robot configuration.
      bool status = hppDevice_->hppSetCurrentConfig (cfg_);
      hppDout (info, "configuration: (" << cfg_ << ")");
      if (!status)
	assert (0 && "Can not set configuration.");

      // Compute distance.
      DistAndPairsOfPointsResult result = distanceProxy_ (cfg_);

      hppDout (info, "pointEnv: (" << result.pointEnv << ")");
      hppDout (info, "pointBody: (" << result.pointBody << ")");

      // If distance is equal to 0, gradient is zero vector
      if (result.distance == 0)
	return;

      // Unit vector linking point on object to point on body
      ublas::vector<double> unit (2);
      unit (0) = result.pointBody[0] - result.pointEnv[0];
      unit (1) = result.pointBody[1] - result.pointEnv[1];
      unit *= 1. / norm_2 (unit);

      hppDout (info, "unit: (" << unit << ")");

      // compute the jacobian of the closest point on the robot
      // with respect to config
      // result is stored in jacobianBody
      CjrlJoint *jrlRootJoint = hppDevice_->getRootJoint ()->jrlJoint ();
      assert (jrlRootJoint);
      CkwsJointShPtr kwsRootJoint =
	hppDevice_->getRootJoint ()->kppJoint ()->kwsJoint ();
      assert (kwsRootJoint);

      CkitMat4 jointPos = kwsRootJoint->currentPosition ();

      hppDout (info, "jointPos: " << jointPos);

      CkcdMat4 invJointPos (jointPos.inverse ()) ;
      CkcdPoint kitPointInJointFrame = invJointPos * result.pointBody;
      vector3d pointInJointFrame;

      hppDout (info, "kitPointInJointFrame: (" << kitPointInJointFrame << ")");

      pointInJointFrame (0) = kitPointInJointFrame[0];
      pointInJointFrame (1) = kitPointInJointFrame[1];
      pointInJointFrame (2) = 0;

      jrlRootJoint->computeJacobianJointWrtConfig ();
      matrix<double> jB (6, hppDevice_->numberDof ());
      jrlRootJoint->getJacobianPointWrtConfig (pointInJointFrame, jB);
      matrix<double> jacobianBody (2, 3);

      for (unsigned int i = 0; i < 2; i++)
	{
	  jacobianBody (i,0) = jB( i,0);
	  jacobianBody (i,1) = jB (i,1);
	  jacobianBody (i,2) = jB (i,5);
	}

      hppDout (info, "jacobianBody: " << jacobianBody);

      // Compute the variation of the configuration with respect to the spline
      // parameters
      dp_->spline ().setParameters (removeScaleFromParameters (p));
      matrix<double> jacobianSpline =
	dp_->spline ().variationConfigWrtParam
	(tpt_.getTime (dp_-> spline ().timeRange ()));

      hppDout (info, "jacobianSpline: " << jacobianSpline);

      subrange (grad, 1, grad.size ()) =
	prod (ublas::vector<double> (prod (unit, jacobianBody)), jacobianSpline);

      hppDout (info, "grad: (" << grad << ")");
      return;
    }
  } // end of namespace walkFootPlanner.
} // end of namespace hpp.
