// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp/walkfootplanner.
//
// hpp/walkfootplanner is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.
//
// hpp/walkfootplanner is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with hpp/walkfootplanner.  If not, see
// <http://www.gnu.org/licenses/>.

#include <iostream>

#include <KineoKCDModel/kppKCDBox.h>
#include <KineoModel/kppLicense.h>
#include <hpp/model/body.hh>
#include <hpp/model/device.hh>
#include <kwsIO/kwsioConfig.h>

#include <hpp/walkfootplanner/planner.hh>

#include "planner-test.hh"

static const unsigned DOF_X = 0;
static const unsigned DOF_Y = 1;
static const unsigned DOF_Z = 2;
static const unsigned DOF_RX = 3;
static const unsigned DOF_RY = 4;
static const unsigned DOF_RZ = 5;

namespace util {
  double prec (const double& x)
  {
    if (fabs (x) < 1e-3) return 0;
    return x;
  }
}

/// \brief Test of class ChppWalkFootPlanner
int PlannerTest::getKineoLicense ()
{
  if (!CkppLicense::initialize ())
    {
      std::cerr << "Error: license for KPP SDK was not found" << std::endl;
      return -1;
    }

  std::cout << "license for KPP SDK was found" << std::endl;
  return 0;
}

int PlannerTest::createEnvAndRobot ()
{
  ChppWalkFootPlanner planner(0.2, 0.2, 0.2);

  // Create a box 0.3 x 0.5 x 1.0
  box_ = planner.createRobotBox (-1., -2.5, 0., 2., 2.5, 1.);
  body_ = KIT_DYNAMIC_PTR_CAST
    (hpp::model::Body, box_->CkppDeviceComponent::rootJoint()->attachedBody ());

  if (!body_)
    return -1;
  planner.addHppProblem (box_, 0.01);

  // Create obstacle
  CkppKCDPolyhedronShPtr obj1 = CkppKCDPolyhedron::create ("obstacle.1");

  obj1->addPoint (CkitPoint3 (-1., -1., 0.)); /* vertex 0 */
  obj1->addPoint (CkitPoint3 ( 1., -1., 0.)); /* vertex 1 */
  obj1->addPoint (CkitPoint3 ( 1.,  1., 0.)); /* vertex 2 */
  obj1->addPoint (CkitPoint3 (-1.,  1., 0.)); /* vertex 3 */
  obj1->addPoint (CkitPoint3 ( 0.,  0., 1.)); /* vertex 4 */

  obj1->addTriangle (0, 1, 4);
  obj1->addTriangle (1, 2, 4);
  obj1->addTriangle (2, 3, 4);
  obj1->addTriangle (3, 0, 4);

  obj1->makeCollisionEntity (CkcdObject::IMMEDIATE_BUILD);
  CkitMat4 position;
  position.translate (20.0, 20.0, 0.0);
  obj1->setAbsolutePosition (position);
  planner.addObstacle (obj1);

  return 0;
}

int
PlannerTest::computeDistance
(double inX, double inY, double inTheta)
{
  CkwsConfig config (box_);

  // X
  config.dofValue (DOF_X, inX);
  // Y
  config.dofValue (DOF_Y, inY);
  // Z
  config.dofValue (DOF_Z, 0);
  // Rx
  config.dofValue (DOF_RX, 0);
  // Ry
  config.dofValue (DOF_RY, 0);
  // Rz
  config.dofValue (DOF_RZ, inTheta);

  // Put box in given configuration
  box_->hppSetCurrentConfig (config);

  std::cout << "config=(" << config << ")" << std::endl;

  // Compute distances and closest points
  CkcdPoint pointBody;
  CkcdPoint pointEnv;
  double distance;

  for (unsigned pairId=0; pairId < body_->nbDistPairs (); pairId++)
    {
      body_->distAndPairsOfPoints
	(pairId, distance, pointBody, pointEnv);
      std::cout << "pairId=" << pairId
		<< ", distance=" << distance;

      if (distance > 0)
	std::cout << ", pointBody=(" << util::prec (pointBody[0])
		  << "," << util::prec (pointBody[1])
		  << "," << util::prec (pointBody[2]) << "), "
		  << ", pointEnv(" << util::prec (pointEnv[0])
		  << "," << util::prec (pointEnv[1])
		  << "," << util::prec (pointEnv[2]) << ").";
      std::cout << std::endl;
    }
  std::cout << std::endl;
  return 0;
}

int
PlannerTest::testDistanceComputation ()
{
  std::cout.unsetf(std::ios::floatfield);            // floatfield not set
  std::cout.precision(6);

  double x=0, y=0, theta=0;
  computeDistance (x, y, theta);

  x=8.0; y=8.0; theta=0;
  computeDistance (x, y, theta);

  x=10.0; y=22.0; theta=-M_PI/6.0;
  computeDistance (x, y, theta);

  x=18.0; y=22.0; theta=-M_PI/6.0;
  computeDistance (x, y, theta);

  x=21.0; y=18.0; theta=M_PI/6.0;
  computeDistance (x, y, theta);

  return 0;
}
